# -*- coding: utf-8 -*-
# Copyright © 2007-2013, All rights reserved. GoodData® Corporation, http://gooddata.com

__author__ = "miroslav.hedl@gooddata.com"
__maintainer__ = __author__


import re
import itertools
import collections
import string

################################################################################
##############################   dictionary  part   ############################
################################################################################
_tuple_cache = {}

def create(data, template, additional_fields=None):
    '''
    Converts nested dictionaries to list of named-tuples with fields
    taken from TEMPLATE.

    See included Yaml file for further details.
    :param dict data: dictionary of nested dictionaries/list
                      - fixed structure is presumed
    :param dict template: dictionary specifying matching fields
                          - useful to set from Yaml
    :param dict additional_fields: smart fields with string.template matching
                                   - can reference other fields from string
    :rtype: list of namedtuple
    :return: list of all fields captured from matching `data` against `template`

    >>> xs = iter_structure(..., ...,
    >>>          additional_fields={'NanaMeme':
    >>>                             "It was ${Name} ($occupation) ${date}"})
    >>> x = xs[0]
    >>> x.Name, x.occupation, x.date, x.NonExists
    'Queeg', 'mainframe', 'TODAY', None
    >>> x.NanaMeme
    'It was Queeg (mainframe) TODAY'
    '''

    m_regex = re.compile(r'''
     ^ \s*                                      # ignore initial whitespaces
     (?:
      (?: [$] (?P<bound> [_a-z] [0-9a-z_]*? ))  # match new bounded field name
      \s*
      (?:
       (?:   / (?P<re_bound> .* )     / )     | # but limit values by regular expression
       (?: - / (?P<re_bound_neg> .* ) / )       # but limit values by negation of regular expression
      )?                                      |
      (?:   / (?P<re> .* )     / )            | # limit values by regex
      (?: - / (?P<re_neg> .* ) / )            | # limit values by regex
      (?: [!] (?P<type> [_a-z] [0-9a-z_]*? )) | # limit values to instance of a type
      (?: \' (?P<literal> .*? [^ \'] ) \' )   | # only equals to this string literal
      (?P<normal> [^  \' $ !] .* )            | # only equals to this string literal
     )
     \s* $                                      # ignore ending whitespaces
     ''', flags=re.I | re.X)

    def m_eq(value, templ_value):
        '''
        Match value with defined template field.
        :param string value: value to match against template matcher
        :param string templ_value: template construct that limits possible values
        $Name       - will capture ANY value and add field to named tuple named 'Name'
        $Name/re/   - will capture ONLY values matching regex `re`, add field 'Name' with matched value
        $Name-/re/  - will capture ONLY values NOT matching regex `re`, add field 'Name' with matched value
        /re/        - will let through only values matching regex `re`, name is NOT bounded
        -/re/       - will let through only values NOT matching regex `re`, name is NOT bounded
        !Integral   - will only matches instances of given type (depends on :py:fn`m_get_type`
        literal     - will only pass values == literal
        '$weird_value' - if in single brackets, it's also literal, must be equal
        '''
        retMatch = None
        retBoundName = None
        m_t = m_regex.match(templ_value)
        if m_t:
            gd = m_t.groupdict()
            if gd.get('bound'):
                retBoundName = gd['bound']
                retMatch = True
            if gd.get('re_bound'):
                retMatch = re.match(gd['re_bound'], value)
            if gd.get('re'):
                retMatch = re.match(gd['re'], value)
            if gd.get('re_bound_neg'):
                retMatch = not re.match(gd['re_bound_neg'], value)
            if gd.get('re_neg'):
                retMatch = not re.match(gd['re_neg'], value)
            if gd.get('type'):
                retMatch = isinstance(value, m_get_type(gd['type']))
            if gd.get('literal'):
                retMatch = value == gd['literal']
            if gd.get('normal'):
                retMatch = value == gd['normal']
        return retMatch, retBoundName

    def iteritems(s):
        '''helper - takes/returns same things as enumerate(), for dict'''
        # return sorted(getattr(s, 'iteritems', s.items)())
        return getattr(s, 'iteritems', s.items)()

    def row_tuple(fields_str, additional_fields=None):
        '''
        Will return proper instance of namedtuple (all named Row).

        :param string fields_str: string of field names to conctruct named tuple from.
        For two same values of `fields_str`, you will always get the same type of namedtuple constructor (therefore isinstance works between fields combinations).
        :param dict additional_fiels: Row tuple will be extended to be asked on arbitrary value - defined in Yaml file. Value can reference other tuples's keys.
        :rtype: constructor for namedtuple defined by `fields_str`
        :return: constructor from cache or newly created one (added to cache right away)
        '''

        def mk__getattr__(additional_fields):
            """
            Extend attribute of `Row` namedtuple.
            Return `None` on invalid field.
            Return formatted string when name is in TEMPLATE dictionary.
            """
            def _attr_f(t, self, name):
                if t and name in t:
                    return string.Template(t[name]).safe_substitute(**self._asdict())
                return None
            return lambda self, name: _attr_f(additional_fields, self, name)

        if fields_str not in _tuple_cache:
            Row = collections.namedtuple('Row', fields_str)
            Row.__getattr__ = mk__getattr__(additional_fields=additional_fields)
            Row.__str = Row.__str__
            Row.__str__ = lambda t: (
                '|<| ' +
                ' | '.join('%s: %s' % (f, getattr(t, f)) for f in t._fields)
                + ' |>|')
            _tuple_cache[fields_str] = Row
        return _tuple_cache[fields_str]

    def is_iterable(s):
        '''Object is iterable but not string.'''
        return hasattr(s, '__iter__')

    def is_scalar(obj):
        '''Object is scalar / atom. Cannot be delved into.'''
        return not (isinstance(obj, collections.Mapping) or is_iterable(obj))

    def iter_tuplepairs(structure):
        '''
        Make sure dictionary and/or list iterate with same protocol.

        :rtype: iterator((value, value))
        :return: iterator of pairs [('key1', 'val1'), ('key1', 'val1'), ...]  or [(1, 'val1'), (2, 'val3')]
        '''
        # is some kind of dictionary
        if isinstance(structure, collections.Mapping):
            return iteritems(structure)
        # is some kind of enumarable, but NOT string
        elif isinstance(structure, collections.Sequence) or is_iterable(structure):
            return enumerate(structure)
        else:
            return tuple()

    def m_get_type(val):
        '''
        Integral type is only type that this function can verify.
        But it's okay, don't really need more types.

        :rval: bool
        :return: will check if `val` is instance of Integral
        '''
        ## or rather use eval() ?  ...later
        tp = ''.join(val.split('!'))
        if tp == 'Integral':
            import numbers
            return numbers.Integral
        return None

    def delve_in_template(templ, name, value):
        '''
        Step down the template dict to the proper sub-branch.
        '''
        if name is not None:
            return templ['$' + name]
        elif value in templ:
            return templ[value]
        types = itertools.ifilter(lambda x: x.startswith('!'),
                                  templ.iterkeys())
        for t in types:
            if isinstance(value, m_get_type(t)):
                return templ[t]
        raise IndexError("(VALUE: %s) or (PATTERN: %s) not in '%s'" % (value, name, templ))

    def m_field_dicts(data_fields, templ):
        '''
        Decision predicate if current captured fields satisfy matching rules and can delve further.

        :param dict data_fields: currently captured fields from input data-structure
        :param dict templ: template fields (for corresponding nested level)
        :rval: tuple (bool, list[(name, val))
        :return: Returns tuple with two elements. First - fields match; Second - list of name/value pairs
        '''
        def fields_only(d):
            '''
            :rval: dictionary
            "return: From input dict, return dict only with fields with atomic values
            '''
            flds = {}
            for field_candidate, value in d.iteritems():
                if is_scalar(value):
                    flds[field_candidate] = value
            return flds

        def satisfy_fields(d, t):
            retVal = tuple()
            template_fields = fields_only(t)
            plain_fields = list(itertools.ifilterfalse(lambda s: s.startswith('$'),
                                                       template_fields.iterkeys()))
            # make sure none of named template fields is missing
            if set(d.iterkeys()).issuperset(set(plain_fields)):
                for f in plain_fields:
                    match, bound_name = m_eq(d[f], t[f])

                    if not match:  # mismatch... skip this branch
                        return None, tuple()

                    if bound_name:
                        retVal = retVal + ((bound_name, d[f]),)
                return True, retVal
            return None, tuple()

        return satisfy_fields(data_fields, templ)

    def m_keyname(key, value, templ):
        def _match_names(name, templ):
            for pattern in templ.iterkeys():
                match, bound_name = m_eq(name, pattern)
                if match:
                    return match, bound_name
            return None, None

        state = []
        tv = templ.get(key)
        if is_scalar(value) or (tv is not None and is_scalar(tv)):
            state.append('field')
        else:
            state.append('nested')

        match, bound_key_name = _match_names(key, templ)
        if match:
            return state, bound_key_name
        return None, None

    def do_iter(path, container, t):
        children = []
        fields = {}

        for k, v in iter_tuplepairs(container):
            state, bound_key_name = m_keyname(k, v, t)
            if state is None:
                continue
            if 'field' in state:
                fields[k] = v
            if 'nested' in state:
                children.append((bound_key_name, k, v))

        match, bound_names = m_field_dicts(fields, t)
        if not match:
            return

        for (name, val, child) in children:
            sub_templ = delve_in_template(t, name, val)
            if isinstance(sub_templ, collections.Mapping):
                do_iter(path + bound_names + ((name, val),),
                        child,
                        sub_templ)
            elif is_iterable(sub_templ):
                for one_of_templ in sub_templ:
                    do_iter(path + bound_names + ((name, val),),
                            child,
                            one_of_templ)

        if len(children) == 0:
            processed.append(path + bound_names)

    processed = []
    do_iter(tuple(), data, template)
    rows = []
    for p in processed:
        traversed = zip(*filter(lambda t: t[0] is not None and t[1], p))
        fields = ' '.join(traversed[0])
        values = traversed[1]
        Row = row_tuple(fields, additional_fields=additional_fields)
        rows.append(Row(*values))
    return rows



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2012, GoodData(R) Corporation. All rights reserved

import os
import sys
from setuptools import setup

DOCDIR = '/usr/share/doc/smoker'
INITDIR = '/etc/rc.d/init.d'

# This package is named gdc-smoker on Pypi (register or upload actions)
if 'upload' in sys.argv or 'register' in sys.argv:
    name = 'gdc-smoker'
else:
    name = 'smoker'

# Parameters for build
params = {
    'name': name,
    'version': '2.0.6',
    'packages': [
        'smoker',
        'smoker.server',
        'smoker.server.plugins',
        'smoker.client',
        'smoker.client.out_junit',
        'smoker.client.plugins',
        'smoker.logger',
        'smoker.util'
        ],
    'scripts': [
        'bin/smokerd.py',
        'bin/smokercli.py',
        'bin/check_smoker_plugin.py',
        ],
    'url': 'https://github.com/gooddata/smoker',
    'download_url': 'https://github.com/gooddata/smoker',
    'license': 'BSD',
    'author': 'GoodData Corporation',
    'author_email': 'python@gooddata.com',
    'maintainer': 'Filip Pytloun',
    'maintainer_email': 'filip@pytloun.cz',
    'description': 'Smoke Testing Framework',
    'long_description': open('smoker/DESCRIPTION.md').read(),
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Monitoring',
    ],
    'platforms': ['POSIX'],
    'provides': ['smoker'],
    'install_requires': ['PyYAML', 'argparse', 'simplejson', 'psutil', 'setproctitle', 'Flask-RESTful'],
    'data_files': [
        (INITDIR, ['rc.d/init.d/smokerd']),
        (DOCDIR, ['etc/smokerd-example.yaml', 'etc/smokercli-example.yaml'])
    ],
    'test_suite': 'tests',
    'package_data': {'smoker': ['DESCRIPTION.md']}
}

setup(**params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2012, GoodData(R) Corporation. All rights reserved

"""
Client tool for GDC Smoker daemon

USAGE
    Action switchers:
      -f, --force           Force plugins run (otherwise just print last results)
      -l, --list            List plugins
      -h, --help            Show this help and exit

    Target host switchers:
      -s HOSTS [HOSTS ...], --hosts HOSTS [HOSTS ...]
                            Hosts with running smokerd (default localhost)

    Filters:
      -p PLUGINS [PLUGINS ...], --plugins PLUGINS [PLUGINS ...]
                            Filter plugins by names, can't be used with --filter
                            option
      --exclude-plugins EXCLUDE_PLUGINS [EXCLUDE_PLUGINS ...]
                            Exclude plugins by names
      --filter FILTER [FILTER ...]
                            Filter plugins by it's parameters, eg. --filter
                            'Category connectors'
      --exclude             Negative filters (exclude) - currently works only for one filter
      --category CATEGORY   Filter plugins by Category parameter
      --component COMPONENT
                            Filter plugins by Component parameter (eg. server)
      --health              Filter plugins by Type healthCheck
      --smoke               Filter plugins by Type smokeTest

      --filter-status FILTER_STATUS [FILTER_STATUS ...]
                            Filter plugins by it's state, eg. --filter-state ERROR WARN
      --nook                Only non-OK plugins (ERROR, WARN, UNKNOWN). Can't be
                            used together with --filter-state
      --error               Only ERROR. Can't be used together with --filter-state
      --warn, --warning     Only WARN. Can't be used together with --filter-state
      --unknown             Only UNKNOWN. Can't be used together with --filter-state
      --ok                  Only OK. Can't be used together with --filter-state

    Output switchers:
      -o PRETTY, --pretty PRETTY
                            Output format: minimal / normal / long / full / raw /
                            json / tap
      --no-colors           Don't use colors in output
      -v, --verbose         Be verbose
      -d, --debug           Debug output

OUTPUT
    Use parameter -o / --pretty to control output
        minimal - host, errored and warned plugins and errored and warned components + error and warning messages
        normal  - host, all plugins, errored and warned components + error and warning messages
        long    - host, all plugins and component results + all messages
        full    - host, all plugins and component results + all messages + plugin info

        raw     - pprint output
        json    - simplejson output

    It's formatted in following scheme
        {format_host}
        for each plugin:
            {format_plugin}
            for each message:
                {format_plugin_run}
                {format_plugin_params}
                {format_plugin_msg}

            for each component:
                {format_plugin_component}
                for each message:
                    {format_plugin_component_msg}

EXAMPLES
    Show results on current host with minimal output
        smoker.py -o minimal

    The same as above with normal output but force immediate run
        smoker.py -f

    Force run of all plugins and print normal output for 2 specific nodes
        smoker.py -f -s server1 server2

    Get results for plugins with category services, component apache
        smoker.py --category services --component apache

    Get results for plugin Uname and Uptime and don't use colors
        smoker.py -p Uname Uptime --no-colors

    Run only smoke tests (no health checks - you should always use --smoke or --health)
        smoker.py --smoke -f

    Run again only smoke tests that failed
        smoker.py --smoke --nook -f

    List plugins that are smoke tests and doesn't name selinux and rpm
        smoker.py --smoke --exclude-plugins selinux rpm --list

    Exclude plugins of category workers
        smoker.py --exclude --category services
"""

import sys
from smoker.client.cli import main

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright © 2007-2013, All rights reserved. GoodData® Corporation, http://gooddata.com

__author__ = "miroslav.hedl@gooddata.com"
__maintainer__ = __author__


YAML_CONFIG = \
'''
AdditionalFields:
  ClassName:
    ${Node}.${Plugin}
  TestSuiteName:
    node $Node
  TestCaseName:
    plugin $Plugin
  ErrorMessage:
    Smoker plugin failed. See STD-ERR for more details.


HtmlTestSuiteAttr:
  errors: _count
  failures: _count
  tests: _count
  name: TestSuiteName
  hostname: Node
  id: UNUSED
  package: UNUSED
  timestamp: LastRun


# this is ignored and hard-coded
#
# HtmlTestCaseElem:
#   system-out: MsgInfo
#   system-err: MsgWarn

HtmlTestCaseAttr:
  classname: ClassName
  name: CaseName
  status: CaseStatus


All:
  '$Node':
    plugins:
      '$Plugin':
        lastResult:
          status: '$PluginStatus /ERROR|WARN|OK/'
          lastRun: '$LastRun'
          messages:
            info:  '$MsgInfo'
            warn:  '$MsgWarn'
            error:  '$MsgError'
          componentResults:
            '!Integral':
             componentResult:
              messages:
                info:  '$MsgInfo'
                warn:  '$MsgWarn'
                error:  '$MsgError'
              name:   '$CaseName'
              status: '$CaseStatus'
'''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2012, GoodData(R) Corporation. All rights reserved

"""
Example plugin that will return
basic informations about system
"""

import os

from smoker.server.plugins import BasePlugin

class Plugin(BasePlugin):
    def run(self):
        # Plugin can access it's parrent
#       self.plugin.params (or get_param() method)
#       self.plugin.result

        uname = os.uname()
        msg = ' '.join(uname)

        self.result.set_status('OK')
        self.result.add_info(msg)

        return self.result



#!/usr/bin/env python
"""
Not mine, found over there:
http://code.activestate.com/recipes/576819-logging-to-console-without-surprises/

Licensed under the MIT License and slightly modified.
"""
import logging
import sys

class LevelHandler(logging.StreamHandler):
    """A handler that logs to console in the sensible way.

    StreamHandler can log to *one of* sys.stdout or sys.stderr.

    It is more sensible to log to sys.stdout by default with only error
    (logging.ERROR and above) messages going to sys.stderr. This is how
    ConsoleHandler behaves.
    """
    def __init__(self, stream_greater_or_equal=sys.stderr,
                 stream_lower=sys.stdout,
                 level=logging.ERROR):
        """
        Initialize handler
        """
        # can't use line bellow, becuase StreamHandler is not
        # a new-style class
        #super(LevelHandler, self).__init__()
        logging.StreamHandler.__init__(self)
        self._level = level
        # keep it set as None
        self.stream = None
        self._stream_lower = stream_lower
        self._stream_greater_or_equal = stream_greater_or_equal

    def emit(self, record):
        """
        Overriden emit.
        """
        if record.levelno >= self._level:
            return self._emit(record, self._stream_greater_or_equal)
        else:
            return self._emit(record, self._stream_lower)

    def _emit(self, record, stream):
        self.stream = stream
        try:
            return logging.StreamHandler.emit(self, record)
            # can't use line bellow, because StreamHandler is not
            # a new-style class
            #return super(LevelHandler, self).emit(self, record)
        except:
            raise
        else:
            self.stream = None

    def flush(self):
        # Workaround a bug in logging module
        # See:
        #   http://bugs.python.org/issue6333
        if self.stream and hasattr(self.stream, 'flush') and not self.stream.closed:
            logging.StreamHandler.flush(self)
